//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CluelessPlus - an noai-AI                                       //
//                                                                  //
//////////////////////////////////////////////////////////////////////
// 
// CluelessPlus is based on original Clueless which was written in 
// the beginning of 2008, when NoAI API was announced.
//
// In contrast to original Clueless, CluelessPlus make use of the
// library pathfinder, so it can play on maps with build on slope
// enabled. Building of stations has been changed so that it try
// many locations in a town using the old algorithm, but without 
// executing any construction. Then afterwards it builds the station
// at the best found location. Clueless original built *lots* of
// stations all over the town and then removed all but the one it 
// wanted to keep. 
//
// Author: Zuu (Leif Linse), user Zuu @ tt-forums.net
// Purpose: To play around with the noai framework.
//               - Not to make the next big thing.
// Copyright: Leif Linse - 2008, 2009
// License: GNU GPL - version 2



require("roadbuilder.nut"); 
require("helper.nut");
require("stationstatistics.nut");


//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: ScoreList                                                //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class ScoreList
{
	list = null;

	sort_max = true;
	sorted = false;

	constructor() {
		list = [];
		sort_max = true;
		sorted = false;
	}

	function Push(value, score);
	function PopMax();
	function PopMin();

	function CompareMin(a, b); // min first
	function CompareMax(a, b); // max first
}

function ScoreList::Push(value, score)
{
	list.append([value, score]);
	sorted = false;
}
function ScoreList::PopMax()
{
	if(list.len() <= 0)
	{
		return null;
	}
	if(!sorted || sort_max)
	{
		list.sort(this.CompareMin);
	}

	return list.pop()[0];
}
function ScoreList::PopMin()
{
	if(list.len() <= 0)
	{
		return null;
	}
	if(!sorted || !sort_max)
	{
		list.sort(this.CompareMax);
	}

	return list.pop()[0];
}
function ScoreList::CompareMin(a, b)
{
	if(a[1] > b[1]) 
		return 1
	else if(a[1] < b[1]) 
		return -1
	return 0;
}
function ScoreList::CompareMax(a, b)
{
	if(a[1] > b[1]) 
		return -1
	else if(a[1] < b[1]) 
		return 1
	return 0;
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: TownFinder                                               //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class TownFinder
{
	// each item in search list is an array of these stuff:
	// [0] => town ID
	// [1] => town population
	search_list = [];

	helper = null;

	conf_min_distance = 20;
	conf_min_population = 150;

	constructor() {
		search_list = [];
		this.helper = Helper();
	}

	function SelectSearchList();
	function FindTwoTownsToConnectByRoad(maxDistance);
	function SelectSearchList_CompareTownByPopulation(a, b);
	function FindTwoTownsToConnectByRoad_CompareTownPairsByScore(a, b);
}

// Creates a sorted list of all towns (best first), and then shrink it to top 50.
// The top 50-list is then stored list in this.search_list.
function TownFinder::SelectSearchList(connection_list)
{
	local town_list = AITownList();
	local list = [];
	local population;

	for(local town_id = town_list.Begin(); town_list.HasNext(); town_id = town_list.Next())
	{
		if(!AITown.IsValidTown(town_id) || this.helper.IsTownInConnectionList(connection_list, town_id))
		{
			//this.helper.Debug("Town no " + town_id + " is not a valid town or already connected, skip it");	
		}
		else
		{
			population = AITown.GetPopulation(town_id);
			if(population >= this.conf_min_population)
			{
				list.append([town_id, population]);
				//this.helper.Debug("appended town to list " + town_id);
			}
		}
	}
	list.sort(this.SelectSearchList_CompareTownByPopulation);
	list.reverse();

	if(list.len() > 50)
	{
		list.resize(50);
	}

	foreach(val in list)
	{
		//this.helper.Debug(val[0] + " (" + AITown.GetName(val[0]) + ") => " + val[1]);
	}

	this.search_list = list;
}

function TownFinder::SelectSearchList_CompareTownByPopulation(a, b)
{
	if(a[1] > b[1]) 
		return 1
	else if(a[1] < b[1]) 
		return -1
	return 0;
}

function TownFinder::FindTwoTownsToConnectByRoad(maxDistance)
{
	//this.helper.Debug("TownFinder::FindTwoTownsToConnectByRoad()");
	local dist_weight = 1.5;
	local pop_weight  = 1.0;

	local list_pairs = ScoreList();
	local i = 0;
	local j = 0;

	//this.helper.Debug(this.search_list.len());
	for(i = 0; i < this.search_list.len(); i++)
	{
		for(j = i+1; j < this.search_list.len(); j++)
		{
			if(i != j)
			{
				local distance = this.helper.TownDistance(this.search_list[i][0], this.search_list[j][0]);
				local total_population = this.search_list[i][1] + this.search_list[j][1];
				local score = total_population * pop_weight + 
						distance * (1 - this.helper.StepFunction(distance - maxDistance))  * dist_weight;
				if( distance > maxDistance || distance < this.conf_min_distance || total_population < this.conf_min_population )
				{
					score = 0;
				}
				else
				{
					list_pairs.Push([this.search_list[i][0], this.search_list[j][0], score, distance, total_population], score);
				}
				//this.helper.Debug("For city " + AITown.GetName(this.search_list[i][0]) + " and " + AITown.GetName(this.search_list[j][0]) + " the score is: " + score + ". The distance is: " + distance + " and the total population is: " + total_population);
			}
		}
	}

	local best_town_pair = list_pairs.PopMax();

	if(best_town_pair == null)
		return null;

	//this.helper.Debug("Best cities are: " + best_town_pair[0] + " and " + best_town_pair[1] + " with the score: " + best_town_pair[2]);
	return [ best_town_pair[0], best_town_pair[1] ];
}

function TownFinder::FindTwoTownsToConnectByRoad_CompareTownPairsByScore(a, b)
{
	if(a[2] > b[2]) 
		return 1
	else if(a[2] < b[2]) 
		return -1
	return 0;
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: Connection - a transport connection                      //
//                                                                  //
//////////////////////////////////////////////////////////////////////

// My aim is that this class should be as general as possible and allow 2> towns, or even cargo transport
class Connection
{
	helper = null;

	cargo = null;
	cargo_type = null;

	// date built
	date_built = null;
	last_vehicle_manage = null;

	// used towns and industries (arrays)
	town     = null;
	industry = null;

	// links to infrastructure and vehicles (arrays)
	station  = null;
	depot    = null;

	last_station_manage = null;
	station_statistics = null; // array of pointers to StationStatistics objects for each station
	max_station_usage = null;

	connection_failed = false;
	
	constructor(helper) {
		this.helper = helper;
		cargo    = null;
		cargo_type = null;
		date_built = null;
		last_vehicle_manage = null;
		town     = [];
		industry = [];
		station  = [null, null];
		depot    = [null, null];
		last_station_manage = null;
		station_statistics = [];
		max_station_usage = 0;
		connection_failed = false;
	}

	function GetTotalPoputaltion();
	function GetTotalDistance(); // only implemented for 2 towns
	function ManageStations()
	function ManageVehicles();
	function SendVehicleForSelling(vehicle_id);

	function FindEngineModelToBuy();
	function BuyVehicles(num_vehicles, engine_id);
	function NumVehiclesToBuy(connection);

	function GetVehicles();
}

function Connection::GetTotalPoputaltion()
{
	local val = null;
	local pop = 0;
	foreach(val in town)
	{
		if(val != null)
		{
			pop += this.AITown().GetPopulation(val);
		}
	}
	return pop;
}
function Connection::GetTotalDistance() 
{
	if(town.len() != 2)
		return 0;

	return this.AIMap().DistanceManhattan( this.AITown().GetLocation(town[0]), this.AITown().GetLocation(town[1]) );
}
function Connection::FindEngineModelToBuy()
{
	// find a bus model to buy
	local bus_list = AIEngineList(AIVehicle.VT_ROAD);

	bus_list.Valuate(AIEngine.GetCargoType)
	bus_list.KeepValue(this.cargo_type); 

	bus_list.Valuate(AIEngine.IsArticulated)
	bus_list.KeepValue(0); 

	bus_list.Valuate(AIEngine.GetCapacity)
	bus_list.KeepTop(1);

	return bus_list.Begin();
}
function Connection::BuyVehicles(num_vehicles, engine_id)
{
	if(station.len() != 2)
	{
		this.helper.Debug("BuyBuses: wrong number of stations");
		return false;
	}
	local cargo = cargo_type;
	local min_reliability = 85;
	local new_bus;
	local i;
	local exec = AIExecMode();

	local existing_vehicles = GetVehicles();
	
	local num_old_vehicles = existing_vehicles.Count();
	local first_buy = num_old_vehicles == 0;
	local share_orders_with = -1;
	if(!first_buy)
		share_orders_with = existing_vehicles.Begin();

	if(engine_id == null || !AIEngine.IsValidEngine(engine_id))
	{
		this.helper.Debug("BuyBuses: failed to find a bus to build");
		return false;
	}

	this.helper.Debug("BuyBuses: engine name: " + AIEngine.GetName(engine_id));

	local new_buses = [];
	
	// Buy buses
	for(i = 0; i < num_vehicles; ++i)
	{
		new_bus = AIVehicle.BuildVehicle(depot[i%2], engine_id);
		if(AIVehicle.IsValidVehicle(new_bus))
		{
			this.helper.Debug("I built a bus");
			new_buses.append(new_bus);
		}
		else if(i == 0)
		{
			local dep = depot[i%2];
			local x = AIMap.GetTileX(dep);
			local y = AIMap.GetTileY(dep);

			AILog.Info("Error string: " + AIError.GetLastErrorString());

			AILog.Warning("Depot is " + dep + ", location: " + x + ", " + y);

			if(!AIMap.IsValidTile(dep))
				AILog.Warning("Depot is invalid tile!");

			if(!AIRoad.IsRoadDepotTile(dep))
				AILog.Warning("No depot!");

			if(!AIEngine.IsValidEngine(engine_id))
				AILog.Warning("Invalid engine!");

			AILog.Warning("Failed to build bus");
			return false; // if no bus have been built, return false
		}
		else
		{
			num_vehicles = i;
			break;
		}

		if(i>0 || !first_buy)
		{
			if(!AIVehicle.IsValidVehicle(share_orders_with))
			{
				Helper.Debug("no vehicle to share vehicle with. " + share_orders_with)
				No_vehicle_to_share_orders_with
			}
			AIOrder.ShareOrders(new_bus, share_orders_with);
		}

		if(first_buy && i == 0)
		{
			AIOrder.AppendOrder(new_bus, depot[0], AIOrder.AIOF_SERVICE_IF_NEEDED);
			AIOrder.AppendOrder(new_bus, station[0], AIOrder.AIOF_NONE); 
			AIOrder.AppendOrder(new_bus, depot[1], AIOrder.AIOF_SERVICE_IF_NEEDED);  
			AIOrder.AppendOrder(new_bus, station[1], AIOrder.AIOF_NONE);

			share_orders_with = new_bus;
		}
	}

	
	for(i = 0; i < new_buses.len(); ++i)
	{
		if(i%2 == 1)
			AIVehicle.SkipToVehicleOrder(new_buses[i], 1);
	}
	
	for(i = 0; i < new_buses.len(); ++i)
	{
		if(i%2 == 1)
			AIController.Sleep(20);

		AIVehicle.StartStopVehicle(new_buses[i]);
	}

	AILog.Info("Built bus => return true");
	return true;
}
function Connection::NumVehiclesToBuy(engine_id)
{
	local population = GetTotalPoputaltion().tofloat();
	local distance = GetTotalDistance().tofloat();
	local capacity = AIEngine.GetCapacity(engine_id);

	this.helper.Debug("NumVehiclesToBuy(): distance between towns: " + distance);
	this.helper.Debug("NumVehiclesToBuy(): total town population:  " + population);
	this.helper.Debug("NumVehiclesToBuy(): capacity:  " + capacity + " cargo_type: " + cargo_type + " engine_id: " + engine_id);

	local num_bus = 1 + max(0, ((population - 200) / capacity / 10).tointeger());
	local extra = distance/capacity/2;
	num_bus += extra;
	this.helper.Debug("NumVehiclesToBuy(): extra:  " + extra);

	num_bus = num_bus.tointeger();

	this.helper.Debug("Buy " + num_bus + " vehicles");
	return num_bus;
}

function Connection::ManageStations()
{
	if(this.connection_failed)
		return;

	local now = AIDate.GetCurrentDate();
	if(this.last_station_manage != null && now - this.last_station_manage < 5)
		return;
	this.last_station_manage = now;

	local max_usage = 0;

	for(local i = 0; i < this.station.len(); ++i)
	{
		local station_tile = this.station[i];
		local station_statistics = this.station_statistics[i];

		station_statistics.ReadStatisticsData();

		local usage = station_statistics.usage.bus.percent_usage;
		if(usage > max_usage)
			max_usage = usage;
	}

	this.max_station_usage = max_usage;
}

function Connection::ManageVehicles()
{
	// Don't manage vehicles for failed connections.
	if(this.connection_failed)
		return;

	this.helper.Debug("Connection::ManageVehicles called for connection: " + 
			AIStation.GetName(AIStation.GetStationID(this.station[0])) + " - " + 
			AIStation.GetName(AIStation.GetStationID(this.station[1])));

	//AISign.BuildSign(this.station[0], "manage");
	//AISign.BuildSign(this.station[1], "manage");

	local now = AIDate.GetCurrentDate();
	if( ((last_vehicle_manage == null && AIDate.GetYear(now) > AIDate.GetYear(date_built)) || 		   // make first manage the year after it was built
			(last_vehicle_manage != null && last_vehicle_manage + AIDate.GetDate(0, 3, 0) < now )) &&  // and then every 3 months
			AIDate.GetMonth(now) > 2)  // but don't make any management on the first two months of the year
	{
		last_vehicle_manage = now;
		this.helper.Debug("Connection::ManageVehicles time to manage vehicles");
		AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
		// wait one year before first manage, and then every year

		local income = 0;
		local veh_list = AIList();

		local connection_vehicles = GetVehicles();
		local veh = null;
		for(veh = connection_vehicles.Begin(); connection_vehicles.HasNext(); veh = connection_vehicles.Next())
		{
			// If the vehicle is to old => sell it. Otherwise if the vehicle is kept, include it in the income checking.
			if(AIVehicle.GetAge(veh) > AIVehicle.GetMaxAge(veh) - 365 * 2)
			{
				SendVehicleForSelling(veh);
			}
			else
			{
				if(AIVehicle.GetAge(veh) > 90) // Only include vehicles older than 90 days in the income stats
					income += AIVehicle.GetProfitThisYear(veh);
				veh_list.AddItem(veh, veh);
			}
		}

		Helper.Debug("connection income: " + income);

		if( (income < 0 || this.max_station_usage > 200) && veh_list.Count() > 1)
		{
			local veh_to_sell = null;

			// sell one vehicle

			// For some particular reason buses are not sent to depot. Also for some particular reason the depots are not added to order list.

			veh_list.Valuate(AIVehicle.GetAgeLeft)
			veh_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // Vehicle with least amount of time left first
			veh_to_sell = veh_list.Begin();
			if(!AIVehicle.IsValidVehicle(veh_to_sell))
			{
				AILog.Error("Can't sell bad vehicle");
				return;
			}

			AILog.Info("Manage vehicles decided to Sell one bus");
			SendVehicleForSelling(veh_to_sell);
		}
		else 
		{
			local min_rating = 100;
			local max_waiting = 0;
			local station = null;
			foreach(station in this.station)
			{
				local rate = AIStation.GetCargoRating(station, cargo_type);
				local waiting = AIStation.GetCargoWaiting(station, cargo_type);
				if(rate < min_rating)
					min_rating = rate;
				if(waiting > max_waiting)
					max_waiting = waiting;
			}

			if(this.max_station_usage < 170 && (min_rating < 40 || max_waiting > 100))
			{
				// Buy a new vehicle
				helper.Debug("Manage vehicles: decided to Buy a new bus");
				local engine = FindEngineModelToBuy();
				local num = 1 + (max_waiting-100) / 100;
				if(num < 1)
					num = 1;
				BuyVehicles(num, engine);
			}
		}
		
	}
	else
	{
		this.helper.Debug("Connection::ManageVehicles: to early to manage");
	}	
}

function Connection::SendVehicleForSelling(vehicle_id)
{
	helper.Debug("Send vehicle " + AIVehicle.GetName(vehicle_id) + " for selling");

	// Unshare & clear orders
	AIOrder.UnshareOrders(vehicle_id);
	while(AIOrder.GetOrderCount(vehicle_id) > 0)
	{
		AIOrder.RemoveOrder(vehicle_id, 0);
	}

	// Send vehicle to specific depot so it don't get lost
	if(!AIOrder.AppendOrder(vehicle_id, depot[0], AIOrder.AIOF_STOP_IN_DEPOT))
	{
		helper.Debug(AIError.GetLastErrorString());
		Failed_to_append_order
	}

	// remove vehicle from vehicle array!!
	//local key = helper.ArrayFind(vehicle, vehicle_id);
	//if(key)
	//	vehicle.remove(key);

	helper.Debug(AIError.GetLastErrorString());
}

function Connection::GetVehicles()
{
	// Return the intersection of the vehicles that stop on station 0 and station 1.
	local veh0 = AIVehicleList_Station(AIStation.GetStationID(station[0]));
	local veh1 = AIVehicleList_Station(AIStation.GetStationID(station[1]));

	veh0.KeepList(veh1);
	local intersect = veh0;

	return intersect;
}

//////////////////////////////////////////////////////////////////////

function GetVehiclesWithoutOrders()
{
	local empty_orders = AIVehicleList();
	empty_orders.Valuate(AIOrder.GetOrderCount);
	empty_orders.KeepValue(0);
	return empty_orders;
}

//////////////////////////////////////////////////////////////////////
//                                                                  //
//  CLASS: CluelessPlus - the AI main class                         //
//                                                                  //
//////////////////////////////////////////////////////////////////////
class CluelessPlus extends AIController {
	stop     = false;

	loaded_from_save = false;

	helper   = null;
	town_finder = null;
	road_builder = null;
	connection_list = [];

	state_build = false;
	state_ai_name = null;

	conf_ai_name = null
	conf_min_balance = 0;

	// All variables should be initiated with their values below!:  (the assigned values above is only there because Squirrel demands it)
	constructor() {
		stop = false;
		loaded_from_save = false;
		helper = Helper();
		town_finder = TownFinder();
		road_builder = RoadBuilder(helper);
		connection_list = [];

		state_build = false;
		state_ai_name = null;

		conf_ai_name = ["Clueless", "Cluemore", "Not so Clueless", "Average IQ", 
				"Almost smart", "Smart", "Even more smart", "Almost intelligent", 
				"Intelligent", "Even more intelligent", "Expert", "Better than expert", 
				"Geek", "Master of universe", "Logistic king", "Ultimate logistics"];
		conf_min_balance = 20000;
	}

	// WARNING: the arguments of the following functions might be wrong. Look at the implementation for the exact arguments. 
	// Squirrel don't complain if the arguments of the definitions and implementation don't match. 
	// Squirrel basically don't look at the definitions below:
	function Start();
	function Stop();

	function Save();
	function Load(version, data);
	function ReadConnectionsFromMap(); // Instead of storing data in the save game, the connections are made out from groups of vehicles that share orders.

	function SetCompanyName(nameArray);

	function ManageLoan();
	function RoundLoanDown(loanAmount); // Helper
	function GetMaxMoney();

	function BuildServiceInTown();
	function BuildBusStopInTown(town);
	function BuyBuses(connection); 

	// if tile is not a road-tile it will search for closest road-tile and then start searching for a location to place it from there.
	function BuildBusStopNextToRoad(tile, min_loops, max_loops); 
	function BuildTruckStopNextToRoad(tile, min_loops, max_loops); 
	function BuildDepotNextToRoad(tile, min_loops, max_loops); 
	function BuildNextToRoad(tile, what, min_loops, max_loops);  // 'what' can be any of: "BUS_STOP", "TRUCK_STOP", "DEPOT". ('what' should be a string)

	function FindClosestRoadTile(tile, max_loops);

	function ConnectByRoad(tile1, tile2, max_loops);
	function FindRoadExtensionTile(road_tile, target_tile, min_loops, max_loops); // road_tile = start search here
	                                                                              // target_tile = when searching for a place to extend existing road, we want to get as close as possible to target_tile
																				  // min_loops = search at least for this amount of loops even if one possible extension place is found (return best found)
																				  // max_loops = maximum search loops before forced return
	function FindRoadExtensionTile_SortByDistanceToTarget(a, b); // Helper

	function ConnectByRoad(tile1, tile2);
}

function CluelessPlus::Start()
{
	this.Sleep(1);

	AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);

	// Company Name
	if(this.loaded_from_save)
	{
		this.state_ai_name = AICompany.GetName(AICompany.COMPANY_SELF);
	}
	else
	{
		this.state_ai_name = this.SetCompanyName(this.conf_ai_name);
	}
	this.helper.company_name = this.state_ai_name; // Debug() need to know the (base) name of the company. 

	// Rebuild the connections structure if loading a save game
	if(this.loaded_from_save)
	{
		helper.Debug("Map loaded => Read connections from the map ...");
		ReadConnectionsFromMap();
		helper.Debug("All connections have been read from the map");
	}

	if(AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_ROAD))
	{
		AILog.Error("Road transport mode is disabled for AIs. This AI is road-only.");
		AILog.Info("Enable road transport mode in advanced settings if you want that this AI should build something");
	}
	
	state_build = false;
	local last_manage_time = AIDate.GetCurrentDate();
	local not_build_info_printed = false;

	local i = 0;
	while(!this.stop)
	{
		i++;
		this.Sleep(1);

		// Sometimes...
		if(i%10 == 1)
		{
			// ... manage our loan
			ManageLoan();

			// Get a list of available buses, so we don't construct if there are no buses
			local engine_list = AIEngineList(AIVehicle.VT_ROAD);
			engine_list.Valuate(AIEngine.GetCargoType)
			local cargo_type = Helper.GetPAXCargo();
			engine_list.KeepValue(cargo_type); 
			engine_list.Valuate(AIEngine.IsArticulated)
			engine_list.KeepValue(0); 

			// Also only manage our vehicles if we have any.
			local bus_list = AIVehicleList();

			// ... check if we can afford to build some stuff
			if(this.GetMaxMoney() > 100000 && !engine_list.IsEmpty() && !AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_ROAD))
			{
				state_build = true;
				not_build_info_printed = false;
			}
			else
			{
				if(!not_build_info_printed)
				{
					not_build_info_printed = true;
					AILog.Info("Not enough money to construct or there is no road engines available (will check every now and then, but only this message is printed to not spam the log)");
				}
			}

			if(!bus_list.IsEmpty() && !AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_ROAD))
			{
				Helper.Debug("Look for buses to sell / send to depot for selling");

				// check if there are any vehicles to sell
				local to_sell_in_depot = AIVehicleList();
				to_sell_in_depot.Valuate(AIVehicle.IsStoppedInDepot);
				to_sell_in_depot.KeepValue(1);
				Helper.Debug("num vehicles stopped in depot: " + to_sell_in_depot.Count());
				for(local i = to_sell_in_depot.Begin(); to_sell_in_depot.HasNext(); i = to_sell_in_depot.Next())
				{
					Helper.Debug("Sell vehicle " + i + ": " + AIVehicle.GetName(i));
					if(!AIVehicle.SellVehicle(i)) // sell
						AILog.Info("Error string: " + AIError.GetLastErrorString());
				}

				// check if there are any vehicles without orders that don't tries to find a depot
				// the new selling code adds a depot order, but there may exist vehicles without orders roaming around for other reasons
				local to_send_to_depot = GetVehiclesWithoutOrders();
				Helper.Debug("num vehicles without orders: " + to_send_to_depot.Count());
				to_send_to_depot.Valuate(AIOrder.IsGotoDepotOrder, AIOrder.ORDER_CURRENT);
				to_send_to_depot.KeepValue(0);
				Helper.Debug("num vehicles does not go to depot: " + to_send_to_depot.Count());
				for(local i = to_send_to_depot.Begin(); to_send_to_depot.HasNext(); i = to_send_to_depot.Next())
				{
					Helper.Debug("Send vehicle " + AIVehicle.GetName(i) + " to depot");
					//AIVehicle.SendVehicleToDepot(i); // send to depot
				}


				// check if we should manage the connections
				local now = AIDate.GetCurrentDate();
				if(now - last_manage_time > AIDate.GetDate(0, 1, 0))
				{
					helper.Debug("Time to manage connections");
					AILog.Warning("Time to manage connections");
					last_manage_time = now;
					local connection = null;
					foreach(connection in connection_list)
					{
						connection.ManageStations();
						connection.ManageVehicles();
					}
				}
			}
		}	

		if(state_build)
		{
			// Scan for two towns to connect
			local towns = null;
			local connection = Connection(helper);
			connection.cargo_type = Helper.GetPAXCargo();
			this.town_finder.SelectSearchList(connection_list);
			towns = this.town_finder.FindTwoTownsToConnectByRoad( 100 ); // Parameter 1: max-distance

			// If we found two towns
			if(towns != null && towns.len() == 2)
			{

				local failed = false;
				this.helper.Debug("Connect " + AITown.GetName(towns[0]) + " with " + AITown.GetName(towns[1]));

				// We don't want to worry about budgeting exactly how much money that is needed, get as much money as possible.
				// We can always pay back later. 
				AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());

				// Build bus-stops
				local bus_stop = [null, null];
				bus_stop[0] =  BuildBusStopInTown(towns[0]);
				this.helper.Debug("bus-stop 1 built");
				bus_stop[1] =  BuildBusStopInTown(towns[1]);
				this.helper.Debug("bus-stop 2 built");

				// Build depots
				local depot = [null, null];
				if(bus_stop[0] && bus_stop[1])
				{
					depot[0] = BuildDepotNextToRoad(bus_stop[0], 0, 100);
					this.helper.Debug("depot 1 built");
					depot[1] = BuildDepotNextToRoad(bus_stop[1], 0, 100);
					this.helper.Debug("depot 2 built");
				}

				// Check that we built all buildings
				if(bus_stop[0] == null || bus_stop[1] == null || depot[0] == null || depot[1] == null)
				{
					failed = true;
				}

				// save town, station and depot in connection data-structure.
				connection.town.append(towns[0]);
				connection.town.append(towns[1]);
				connection.station[0] = bus_stop[0];
				connection.station[1] = bus_stop[1];
				if(!failed)
				{
					connection.station_statistics.append(StationStatistics(AIStation.GetStationID(connection.station[0])));
					connection.station_statistics.append(StationStatistics(AIStation.GetStationID(connection.station[1])));
				}
				connection.depot[0] = depot[0];
				connection.depot[1] = depot[1];
				
				local connected = false;
				this.helper.Debug("bus-stops built");
				if(!failed)
					connected = ConnectByRoad(AIRoad.GetRoadStationFrontTile(bus_stop[0]), AIRoad.GetRoadStationFrontTile(bus_stop[1]), 900) 


				// Only buy buses if we actually did connect the two cities.
				if(connected && !failed)
				{	
					this.helper.Debug("connected by road");
					// BuyBuses save the IDs of the bough buses in the connection data-structure
					BuyBuses(connection); 
					this.helper.Debug("bough buses");
				}
				else
				{
					this.helper.Debug("failed to connect by road");

					if(connection.depot[0])
						AITile.DemolishTile(connection.depot[0]);
					if(connection.depot[1])
						AITile.DemolishTile(connection.depot[1]);
					if(connection.station[0])
						AITile.DemolishTile(connection.station[0]);
					if(connection.station[1])
						AITile.DemolishTile(connection.station[1]);

					connection.connection_failed = true;			// store that this connection faild so we don't waste our money on buying buses for it.
				}

				// Store the connection so we don't build it again.
				connection.date_built = AIDate.GetCurrentDate();
				connection_list.append(connection); 

				state_build = false;
			}
			else
			{
				//this.helper.Debug("Could not find two towns to connect");
			}
		}

		// Pay back unused money
		ManageLoan();
	}
}
function CluelessPlus::Stop()
{
	this.helper.Debug("CluelessPlus::Stop()");
	this.stop = true;
}

function CluelessPlus::Save()
{
	AILog.Warning("CluelessPlus do not support save/load!");
	local table = {};
	return table;
}

function CluelessPlus::Load(version, data)
{
	// CluelessPlus does not support save/load, so kill ourself if 
	// a user tries to load a savegame with CluelessPlus as AI.
	this.loaded_from_save = true;
	AILog.Warning("Loading..");
	AILog.Info("Previously saved with AI version " + version);
}

function CargoOfVehicleValuator(vehicle_id)
{
	local engine_id = AIVehicle.GetEngineType(vehicle_id);
	return AIEngine.GetCargoType(engine_id);
}

function CluelessPlus::ReadConnectionsFromMap()
{
	// Get all road vehicles that carries passengers => buses
	local uncategorized_vehicles = AIVehicleList();
	uncategorized_vehicles.Valuate(AIVehicle.GetVehicleType);
	uncategorized_vehicles.KeepValue(AIVehicle.VT_ROAD);
	uncategorized_vehicles.Valuate(CargoOfVehicleValuator);
	uncategorized_vehicles.KeepValue(Helper.GetPAXCargo());

	local unused_stations = AIStationList(AIStation.STATION_BUS_STOP);
	unused_stations.Valuate(Helper.ItemValuator);

	while(uncategorized_vehicles.Count() > 0)
	{
		local veh_id = uncategorized_vehicles.Begin();
		local group = AIVehicleList_SharedOrders(veh_id);

		// remove the vehicles that belongs to the found group from the list of uncategorised vehicles
		uncategorized_vehicles.RemoveList(group);

		// Construct the connection object and read everything needed for the connection from the map.
		helper.Debug("Found connection with " + group.Count() + " vehicles");
		local connection = Connection(helper);
		connection.cargo_type = Helper.GetPAXCargo();

		connection.station = [];
		connection.depot = [];
		for(local i_order = 0; i_order < AIOrder.GetOrderCount(veh_id); ++i_order)
		{
			if(AIOrder.IsGotoStationOrder(veh_id, i_order))
			{
				local station_id = AIStation.GetStationID(AIOrder.GetOrderDestination(veh_id, i_order));
				helper.Debug("Added station: " + AIStation.GetName(station_id));
				connection.station.append(AIOrder.GetOrderDestination(veh_id, i_order));
				connection.station_statistics.append(StationStatistics(station_id));

				// remove station from unused stations list
				unused_stations.RemoveValue(station_id);
			}

			if(AIOrder.IsGotoDepotOrder(veh_id, i_order))
			{
				connection.depot.append(AIOrder.GetOrderDestination(veh_id, i_order));
			}
		}

		// Ignore vehicles with != 2 stations
		if(connection.station.len() != 2)
			continue;

		local station_tile = null;
		connection.town = [];
		foreach(station_tile in connection.station)
		{
			local town_id = AITile.GetClosestTown(station_tile);
			connection.town.append(town_id);
		}

		group.Valuate(AIVehicle.GetAge);
		group.Sort(AIAbstractList.SORT_BY_VALUE, false); // oldest first
		local estimated_construction_date = AIDate.GetCurrentDate() - AIVehicle.GetAge(group.Begin());
		
		connection.date_built = estimated_construction_date;

		connection_list.append(connection);
	}

	// Destroy all unused stations so they don't cost money
	for(local station_id = unused_stations.Begin(); unused_stations.HasNext(); station_id = unused_stations.Next())
	{
		local tile_list = AITileList_StationType(station_id, AIStation.STATION_BUS_STOP);

		for(local tile_id = tile_list.Begin(); tile_list.HasNext(); tile_id = tile_list.Next())
		{
			AITile.DemolishTile(tile_id);
		}
	}
}

// RETURN: new name
function CluelessPlus::SetCompanyName(nameArray)
{
	AICompany.SetPresidentName("Dr. Clue");

	local i = 0;
	//this.helper.Debug("i, before loop = " + i);
	while(i < nameArray.len() && !AICompany.SetName(nameArray[i]))
	{
		this.helper.Debug("i++ = " + i);
		i = i + 1;
	}
	//this.helper.Debug("company name done");
	//this.helper.Debug("i, after loop = " + i);
	return nameArray[i];
}

function CluelessPlus::ManageLoan()
{
	// local constants. ( I've not found how to declare constants in Squirrel yet :( )
	local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	local loan = AICompany.GetLoanAmount();
	local max_balance = this.conf_min_balance + 2 * AICompany.GetLoanInterval(); // Attention: max_balance is the maximum balance we accept before paying back loan.


	if( balance < this.conf_min_balance ) // bigger loan
	{
		//this.helper.Debug("ManageLoan: Want to loan more");
		//this.helper.Debug("Company: " + AICompany.GetCompanyName());

		local new_loan = loan + (this.conf_min_balance - balance) + AICompany.GetLoanInterval();
		if(new_loan < AICompany.GetMaxLoanAmount())
			new_loan = AICompany.GetMaxLoanAmount();

		new_loan = RoundLoanDown(new_loan);
		if(!AICompany.SetLoanAmount(new_loan))
		{
			//this.helper.Debug(this.state_ai_name + " Failed to increase loan amount");
		}
	}
	
	else if( balance > max_balance && loan > 0) // smaller loan
	{
		//this.helper.Debug("ManageLoan: Want to pay back");
		//this.helper.Debug("Company: " + AICompany.GetCompanyName());
		//this.helper.Debug("balance: " + balance);
		//this.helper.Debug("loan: " + loan);
		//this.helper.Debug("max_balance: " + max_balance);

		local pay_back = balance - max_balance;
		local new_loan = RoundLoanDown(loan - pay_back);

		//this.helper.Debug("pay_back: " + pay_back);
		//this.helper.Debug("new_loan: " + new_loan);
		
		if(!AICompany.SetLoanAmount(new_loan))
		{
			//this.helper.Debug(this.state_ai_name + " Failed to decrease loan amount");
		}
			
		//this.helper.Debug("Successfully paid back loan");
	}
}

function CluelessPlus::RoundLoanDown(loanAmount)
{
	return loanAmount - loanAmount % AICompany.GetLoanInterval();
}
function CluelessPlus::GetMaxMoney()
{
	local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	local loan = AICompany.GetLoanAmount();

	local max_possible_balance = balance + (AICompany.GetMaxLoanAmount() - loan);

	return max_possible_balance;
}

function CluelessPlus::BuildBusStopInTown(town) // this function is ugly and should ideally be removed
{
	this.helper.Debug("CluelessPlus::BuildBusStopInTown(" + AITown.GetName(town) + ")");

	if(AITown.IsValidTown(town))
		AILog.Info("Town is valid");
	else
		AILog.Warning("Town is NOT valid");
	
	local location = AITown.GetLocation(town);

	if(!AIMap.IsValidTile(location))
	{
		AILog.Error("Invalid location!");
		return false;
	}

	return BuildBusStopNextToRoad(location, 50, 100 + AITown.GetPopulation(town) / 70);

/*
	local tiles = AITileList();
	tiles.AddRectangle(location - AIMap.GetTileIndex(15, 15), location + AIMap.GetTileIndex(15, 15));
	tiles.Valuate(GetNeighbourRoadCount);
	tiles.KeepAboveValue(0);
	tiles.Valuate(IsWithinTownInfluence, town);
	tiles.KeepValue(true); 
	tiles.Valuate(*/
}

// Build*NextToRoad functions:
// 1. Checks if 'tile' is road
// 2. If not: it locates nearby road using FindClosestRoadTile
// 3. From the starting-point on the road network, search for a suitable 
//    location to build the building. 
//    - it prefers to build the building near the starting-point
function CluelessPlus::BuildBusStopNextToRoad(tile, min_loops, max_loops)
{
	return BuildNextToRoad(tile, "BUS_STOP", min_loops, max_loops, 1);
}
function CluelessPlus::BuildTruckStopNextToRoad(tile, min_loops, max_loops)
{
	return BuildNextToRoad(tile, "TRUCK_STOP", min_loops, max_loops, 1);
}
function CluelessPlus::BuildDepotNextToRoad(tile, min_loops, max_loops)
{
	return BuildNextToRoad(tile, "DEPOT", min_loops, max_loops, 1);
}
function CluelessPlus::BuildNextToRoad(tile, what, min_loops, max_loops, try_number)
{
	local start_tile = tile;

	local green_list = ScoreList();
	local red_list = [];

	local found_locations = ScoreList();

	local i = 0;
	local ix, iy;
	local curr_x, curr_y;
	local curr_tile;
	local adjacent_x, adjacent_y;
	local adjacent_tile;

	local adjacent_loop = [ [-1,0], [1,0], [0,-1], [0,1] ];

	if(!AIMap.IsValidTile(start_tile))
	{
		AILog.Error("Invalid start_tile!");
		return null;
	}

	if(!AIRoad.IsRoadTile(start_tile))
	{
		start_tile = FindClosestRoadTile(start_tile, 4);
		if(!start_tile)
		{
			AILog.Error("failed to find road tile as start_tile was not a road tile!");
			return null;
		}
	}

	curr_tile = start_tile;

	while( i++ < max_loops )
	{
		/*{
			local exec = AIExecMode(); 
			AISign.BuildSign(curr_tile, i+"/"+AIMap.DistanceSquare(curr_tile, start_tile));
		}*/
		local testmode = AITestMode();
		curr_x = AIMap.GetTileX(curr_tile);
		curr_y = AIMap.GetTileY(curr_tile);

		// if we are on a bridge end, add the tile next to other end to green list if it's not in red_list and is accessible from bridge.
		if(AIBridge.IsBridgeTile(curr_tile) || AITunnel.IsTunnelTile(curr_tile))
		{
			local exec = AIExecMode();	

			//AISign.BuildSign(curr_tile, "bridge end");
			local other_end = null;
			if(AIBridge.IsBridgeTile(curr_tile))
			{
				other_end = AIBridge.GetOtherBridgeEnd(curr_tile);
			}
			else 
			{
				other_end = AITunnel.GetOtherTunnelEnd(curr_tile);
			}
			//AISign.BuildSign(other_end, "other end");

			// Get tile next to bridge/tunnel on the other end
			local next_to_other_end = null;
			local x = AIMap.GetTileX(curr_tile) - AIMap.GetTileX(other_end);
			local y = AIMap.GetTileY(curr_tile) - AIMap.GetTileY(other_end);
			if(x != 0)
				x = x / abs(x);
			if(y != 0)
				y = y / abs(y);
			next_to_other_end = other_end - AIMap.GetTileIndex(x, y);
			//AISign.BuildSign(next_to_other_end, "next to other end");
			
			// Add the tile next_to_other_end to green list if it is not in red list and is accessible from the bridge
			if( this.helper.ArrayFind(red_list, next_to_other_end) == null )
			{
				//local test = AITestMode(); // < let's add a road bit at the other end if there is no so we can use that for building next to if needed.
				if(AIRoad.AreRoadTilesConnected(other_end, next_to_other_end) || AIRoad.BuildRoad(other_end, next_to_other_end))
				{
					green_list.Push(next_to_other_end, AIMap.DistanceSquare(next_to_other_end, start_tile));
				}
			}
			
		}
		else
		{
			// scan adjacent tiles

			foreach(adjacent_offset in adjacent_loop)
			{
				adjacent_x = curr_x + adjacent_offset[0];
				adjacent_y = curr_y + adjacent_offset[1];

				adjacent_tile = AIMap.GetTileIndex(adjacent_x, adjacent_y);

				if(!AIMap.IsValidTile(adjacent_tile))
				{
					AILog.Warning("Adjacent tile is not valid");
				}

				if(AIRoad.AreRoadTilesConnected(curr_tile, adjacent_tile))
				{
					if( this.helper.ArrayFind(red_list, adjacent_tile) == null )
					{
						local exec = AIExecMode();	
						green_list.Push(adjacent_tile, AIMap.DistanceSquare(adjacent_tile, start_tile));
						//AISign.BuildSign(adjacent_tile, i+":"+AIMap.DistanceSquare(adjacent_tile, start_tile));
					}
				}
				else if(AIRoad.BuildRoad(adjacent_tile, curr_tile))
				{
					/*{
						local exec = AIExecMode(); 
						AISign.BuildSign(adjacent_tile, i+"|"+AIMap.DistanceSquare(adjacent_tile, start_tile));
					}*/

					local ret;
					

					if(what == "BUS_STOP")
					{
						ret = AIRoad.BuildRoadStation(adjacent_tile, curr_tile, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT);
					}
					else if(what == "TRUCK_STOP")
					{
						ret = AIRoad.BuildRoadStation(adjacent_tile, curr_tile, AIRoad.ROADVEHTYPE_TRUCK, AIStation.STATION_JOIN_ADJACENT);
					}
					else if(what == "DEPOT")
					{
						ret = AIRoad.BuildRoadDepot(adjacent_tile, curr_tile)
					}
					else
					{
						this.helper.Debug("ERROR: Invalid value of argument 'what' to function BuildNextToRoad(tile, what)");
						AIRoad.RemoveRoad(adjacent_tile, curr_tile);
						return null;
					}

					if(ret)
					{
						found_locations.Push([adjacent_tile, curr_tile], AIMap.DistanceSquare(adjacent_tile, start_tile));
					}
				}
				else
				{
					//local exec = AIExecMode();	
					//AISign.BuildSign(adjacent_tile, "x");
				}

				if(i%10 == 0)
				{
					this.Sleep(1);
				}
			}
		}

		// if found at least one location and we have looped at least min_loops. => don't search more
		if(found_locations.list.len() > 0 && i >= min_loops)
		{
			break;
		}

		red_list.append(curr_tile);

		// select best tile from green_list
		curr_tile = green_list.PopMin();

		if(!curr_tile || !AIMap.IsValidTile(curr_tile)) // if green_list was empty
		{
			AILog.Warning("Green list empty in BuildNextToRoad function.");
			break;
		}
	}

	// get best built building
	local best_location = found_locations.PopMin();
	if(best_location == null) // return null, if no location at all was found.
	{
		this.helper.Debug("BuildNextToRoad: failed to build: " + what);
		return null;
	}

	// Build best station
	local road_tile = best_location[1];
	local station_tile = best_location[0];
	
	local ret = false;

	AIRoad.BuildRoad(road_tile, station_tile);
	if(what == "BUS_STOP")
	{
		ret = AIRoad.BuildRoadStation(station_tile, road_tile, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT);
	}
	else if(what == "TRUCK_STOP")
	{
		ret = AIRoad.BuildRoadStation(station_tile, road_tile, AIRoad.ROADVEHTYPE_TRUCK, AIStation.STATION_JOIN_ADJACENT);
	}
	else if(what == "DEPOT")
	{
		ret = AIRoad.BuildRoadDepot(station_tile, road_tile)
	}
	else
	{
		this.helper.Debug("ERROR: Invalid value of argument 'what' to function BuildNextToRoad(tile, what) at constructing on the best tile");
		return null;
	}

	if(!ret)
	{
		// someone built something on the tile so now it's not possible to build there. :( 
		// make another try a few times and then give up
		if(try_number <= 5)
		{
			AILog.Info("BuildNextToRoad retries by calling itself");
			return BuildNextToRoad(tile, what, min_loops, max_loops, try_number+1);
		}
		else
			return null;
	}

	this.Sleep(10);
	//helper.ClearAllSigns();
	
	// return the location of the built station
	return station_tile;
	
}

function CluelessPlus::FindClosestRoadTile(tile, max_radius)
{
	if(!tile || !AIMap.IsValidTile(tile))
		return null;

	if(AIRoad.IsRoadTile(tile))
		return tile;
	
	local r; // current radius

	local start_x = AIMap.GetTileX(tile);
	local start_y = AIMap.GetTileY(tile);

	local x0, x1, y0, y1;
	local ix, iy;
	local test_tile;

	for(r = 1; r < max_radius; ++r)
	{
		y0 = start_y - r;
		y1 = start_y + r;
		for(ix = start_x - r; ix <= start_x + r; ++ix)
		{
			test_tile = AIMap.GetTileIndex(ix, y0)
			if(test_tile != null && AIRoad.IsRoadTile(test_tile))
				return test_tile;

			test_tile = AIMap.GetTileIndex(ix, y1)
			if(test_tile != null && AIRoad.IsRoadTile(test_tile))
				return test_tile;
		}

		x0 = start_x - r;
		x1 = start_x + r;
		for(iy = start_y - r + 1; iy <= start_y + r - 1; ++iy)
		{
			test_tile = AIMap.GetTileIndex(x0, iy)
			if(test_tile != null && AIRoad.IsRoadTile(test_tile))
				return test_tile;

			test_tile = AIMap.GetTileIndex(x1, iy)
			if(test_tile != null && AIRoad.IsRoadTile(test_tile))
				return test_tile;

		}
	}

	return null;
}

function CluelessPlus::FindRoadExtensionTile(road_tile, target_tile, min_loops, max_loops)
{
	// road_tile belong to a road-network of one or more tiles.
	// This function aims to find a tile in this network from which the network can be extended.
	// It stops the search when either:
	// * the whole network have been scanned
	// * it have looped for min_loops times AND have found at least one solution
	// * it have looped for max_loops times
	//
	// Returns the best solution-tile.


	if(road_tile == null || !AIRoad.IsRoadTile(road_tile) || target_tile == null || !AIMap.IsValidTile(target_tile))
		return null;


	local start = road_tile;

	local green_list = ScoreList()
	local red_list = [];

	// TODO: Use ScoreList instead of a plain array.
	local extend_list = []; // each item is an array: [tile, MH distance to target]

	local i = 0;
	local ix, iy;
	local curr_x, curr_y;
	local curr_tile;
	local adjacent_x, adjacent_y;
	local adjacent_tile;

	local adjacent_loop = [ [-1,0], [1,0], [0,-1], [0,1] ];

	curr_tile = start;

	while( i++ < max_loops )
	{
		curr_x = AIMap.GetTileX(curr_tile);
		curr_y = AIMap.GetTileY(curr_tile);

		//AISign.BuildSign(curr_tile, "i:" + i)
		
		// scan adjacent tiles

		foreach(adjacent_offset in adjacent_loop)
		{
			adjacent_x = curr_x + adjacent_offset[0];
			adjacent_y = curr_y + adjacent_offset[1];

			adjacent_tile = AIMap.GetTileIndex(adjacent_x, adjacent_y);

			if(AIRoad.AreRoadTilesConnected(curr_tile, adjacent_tile))
			{
				// special case: we have reached the target-tile.
				if(adjacent_tile == target_tile)
				{
					return adjacent_tile;
				}

				// add tile to green_list if it is not in red_list
				if( this.helper.ArrayFind(red_list, adjacent_tile) == null )
				{
					green_list.Push(adjacent_tile, AIMap.DistanceManhattan(adjacent_tile, target_tile));
				}
			}
			else 
			{
				//this.helper.Debug("CluelessPlus::FindRoadExtensionTile: found one solution");
				// Perhaps the road can be extended here. 
				// Test if it is possible to do so.
				local test_mode = AITestMode();
				if(AIRoad.BuildRoad(curr_tile, adjacent_tile))
				{
					//this.helper.Debug("Possible to build road");
					extend_list.append( [ curr_tile, AIMap.DistanceManhattan(curr_tile, target_tile) ] );
				}
			}

			this.Sleep(1);
		}

		// stop loop if at least one solution found and i > min_loops
		if(extend_list.len() > 0 && i > min_loops)
		{
			break;
		}

		red_list.append(curr_tile);

		// select best tile from green_list
		curr_tile = green_list.PopMin();

		if(!curr_tile || !AIMap.IsValidTile(curr_tile)) // if green_list was empty
		{
			break;
		}
	}

	if(extend_list.len() == 0)
	{
		this.helper.Debug("CluelessPlus::FindRoadExtensionTile: found zero solutions");
		return null;
	}
	else if(extend_list.len() == 1)
		return extend_list[0][0];
	else
	{
		extend_list.sort(FindRoadExtensionTile_SortByDistanceToTarget);
		return extend_list[0][0];
	}

	return null;
}
function CluelessPlus::FindRoadExtensionTile_SortByDistanceToTarget(a, b)
{
	if(a[1] > b[1]) 
		return 1
	else if(a[1] < b[1]) 
		return -1
	return 0;
}

function CluelessPlus::ConnectByRoad(tile1, tile2, max_loops)
{
	return RoadBuilder.ConnectTiles(tile1, tile2, 1);


	if(tile1 == null || tile2 == null || max_loops == null)
		return false;

	if(!AIRoad.IsRoadTile(tile1) || !AIRoad.IsRoadTile(tile2))
		return false;
	
	local start_tile = [null, null];

	/*AISign.BuildSign(tile1, "tile1");
	AISign.BuildSign(tile2, "tile2");*/

	// Assuming that tile1, tile2 might be in the middle of a city, we make
	// some scanning for good road-tiles in the 'border' of the road network 
	// that tile1, tile2 might be connected to.
	//
	// Update: Now we actually trying to find a way to go from town 0 to town 1
	//         using existing road. If that fails we get an extension-point of
	//         the second town too and continue with building road.
	start_tile[0] = FindRoadExtensionTile(tile1, tile2, 300, 400);  // scan from,  scan to,  min loops,  max loops
	if(start_tile[0] == tile2)
		return true; // return true if the other city was found.

	start_tile[1] = FindRoadExtensionTile(tile2, tile1, 40, 100);
	if(start_tile[1] == tile1)
		return true; // return true if the other city was found.


	// Check that the two start-tiles are valid.
	if(!start_tile[0] || !AIMap.IsValidTile(start_tile[0]) || !start_tile[1] || !AIMap.IsValidTile(start_tile[1]))
		return false;
	

	// Now we have two tiles, start_tile[0] and start_tile[1] to connect by road, and both have 
	// adjacent tiles which we can extend into. (The FindRoadExtensionTile-function make sure 
	// that there are adjacent tiles that we can extend to.)

	// The basic idea is to build from curr_tile[0] to curr_tile[1]. There is no green/red-list.
	// It will just select a direction to head at, and then continue until it hits something. 
	// Then it will try the other direction that heads towards curr_tile[1]. If that also fails,
	// it have some minor hacks for minor obstacles. Bigger things like industries can make the 
	// algorithm fail. 
	
	local curr_tile =  start_tile;
	local i = 0;
	local dirr_x = 1;
	local dirr_y = 0;
	local go_x = true;
	local diff_x = null;
	local diff_y = null;
	local target_x = null;
	local target_y = null;
	local next_tile = null;
	local built_road = false;
	local fails = 0;

	diff_x = AIMap.GetTileX(curr_tile[1]) - AIMap.GetTileX(curr_tile[0]);
	diff_y = AIMap.GetTileY(curr_tile[1]) - AIMap.GetTileY(curr_tile[0]);

	if(diff_x > 0)
		dirr_x = 1;
	else
		dirr_x = -1;

	if(diff_y > 0)
		dirr_y = 1;
	else
		dirr_y = -1;

	target_x = AIMap.GetTileX(curr_tile[1]);
	target_y = AIMap.GetTileY(curr_tile[1]);
	if(abs(diff_x) > abs(diff_y))
		go_x = true;
	else
		go_x = false;


	while(curr_tile[0] != curr_tile[1] && ++i <= max_loops)
	{
		if(go_x)
		{
			//this.helper.Debug("ConnectByRoad: go_x");
			next_tile = AIMap.GetTileIndex(AIMap.GetTileX(curr_tile[0]) + dirr_x, AIMap.GetTileY(curr_tile[0]));

			built_road = AIRoad.BuildRoad(curr_tile[0], next_tile);

			//if(built_road)
			if(built_road || (fails < 1 && AIRoad.AreRoadTilesConnected(curr_tile[0], next_tile)) )
			{
				/*if(!built_road)
					AISign.BuildSign(next_tile, "jump");
				else
					AISign.BuildSign(curr_tile[0], "ok");*/
				curr_tile[0] = next_tile;
				fails = 0;
			}
			else
			{
				//AISign.BuildSign(curr_tile[0], "x");
				fails++;
				if(fails == 1)
					go_x = false;
				else
				{
					// try build one step in wrong direction
					next_tile = AIMap.GetTileIndex(AIMap.GetTileX(curr_tile[0]) - dirr_x, AIMap.GetTileY(curr_tile[0]));
					if(AIRoad.BuildRoad(curr_tile[0], next_tile))
					{
						curr_tile[0] = next_tile;
						go_x = false;

						// don't reset 'fails', a non-fail is building in the right direction  (if you don't understand this comment, read the previous one too!)
					}
					else
						break; // bail out
				}
			}
			if((dirr_x == 1 && AIMap.GetTileX(next_tile) >= target_x) || (dirr_x == -1 && AIMap.GetTileX(next_tile) <= target_x))
			{
				go_x = false;
			}
		}
		else if(!go_x)  // must be else, otherwise we might miss the target!
		{
			//this.helper.Debug("ConnectByRoad: !go_x");
			next_tile = AIMap.GetTileIndex(AIMap.GetTileX(curr_tile[0]), AIMap.GetTileY(curr_tile[0]) + dirr_y);

			built_road = AIRoad.BuildRoad(curr_tile[0], next_tile);

			//if(built_road)
			if(built_road || (fails < 1 && AIRoad.AreRoadTilesConnected(curr_tile[0], next_tile)) )
			{
				/*if(!built_road)
					AISign.BuildSign(next_tile, "jump");
				else
					AISign.BuildSign(curr_tile[0], "ok");*/
				curr_tile[0] = next_tile;
				fails = 0;
			}
			else
			{
				//AISign.BuildSign(curr_tile[0], "x");
				fails++;
				if(fails == 1)
					go_x = true;
				else
				{
					// try build one step in wrong direction
					next_tile = AIMap.GetTileIndex(AIMap.GetTileX(curr_tile[0]), AIMap.GetTileY(curr_tile[0]) - dirr_y);
					if(AIRoad.BuildRoad(curr_tile[0], next_tile))
					{
						curr_tile[0] = next_tile;
						go_x = true;
						// don't reset fails, a non-fail is building in the right direction
					}
					else
						break; // bail out
				}
			}
			if((dirr_y == 1 && AIMap.GetTileY(next_tile) >= target_y) || (dirr_y == -1 && AIMap.GetTileY(next_tile) <= target_y))
			{
				go_x = true;
			}
		}

		// Update dirr_*. I'm not sure if it is good to do so, the rest of the code (above) was written as if dirr_* are constant.
		// begin
		diff_x = AIMap.GetTileX(curr_tile[1]) - AIMap.GetTileX(curr_tile[0]);
		diff_y = AIMap.GetTileY(curr_tile[1]) - AIMap.GetTileY(curr_tile[0]);

		if(diff_x > 0)
			dirr_x = 1;
		else
			dirr_x = -1;

		if(diff_y > 0)
			dirr_y = 1;
		else
			dirr_y = -1;
		// end
	}

	if(curr_tile[0] == curr_tile[1])
		return true;

	return false;

}

function CluelessPlus::BuyBuses(connection)
{
	local engine = connection.FindEngineModelToBuy();
	local num = connection.NumVehiclesToBuy(engine);

	AILog.Info("buy engine " + engine);

	connection.BuyVehicles(num, engine);

	return connection;
}

